import io
import os
import pathlib

from .objects import User, FileSize
from .http import Request

class WantsToParty:
    """The base class for all wants-to.party API interactions."""
    def __init__(
        self,
        api_key: str
    ):
        self.data = {"key": api_key}

    def upload_from_bytes(
        self,
        file: io.BufferedIOBase,
        filetype: str,
        *,
        max_bytes: FileSize = None,
        code_length: int = None,
        extension: bool = None,
        custom_code: str = None
    ) -> str:
        """Uploads a file-like object using the raw binary data
        provided.
        Returns the URL of the uploaded file.
        
        Attributes
        -----------
        filepath - The raw binary data associated with the file.
               Using `io.BytesIO` is recommended.
            
        filetype - The file type associated with the file
                   you're uploading. I.E., ".png".

        max_bytes - A maximum number of bytes uploadable. Use
                    the Size class to specify the number of bytes. 
        
        code_length - The length of the unique code to generate.
                      Must be between 5 and 50 characters.

        extension - If `False`, this will not append the file type
                    at the end of the file.
        
        custom_code - The custom file name which will be displayed on
                      wants-to.party.
        """ 
        return Request._post_from_bytes(
            file,
            filetype,
            data=self.data,
            url_args={"max_bytes": max_bytes, "code_length": code_length, "custom_code": custom_code, "extension": extension},
            max_bytes=max_bytes
        )
    
    def upload_from_file(
        self,
        filepath: str,
        *,
        max_bytes: FileSize = None,
        code_length: int = None,
        extension: bool = None,
        custom_code: str = None
    ) -> str:
        """Uploads a local file using the file path provided.
        
        Returns the URL of the uploaded file.
        
        Attributes
        -----------
        file - The path to the file stored on your local disk.
        
        max_bytes - A maximum number of bytes uploadable. Use
        the Size class to specify the number of bytes. 
        
        code_length - The length of the unique code to generate.
                      Must be between 5 and 50 characters.

        extension - If `False`, this will not append the file type
                    at the end of the file.
        
        custom_code - The custom file name which will be displayed on
                      wants-to.party.
        """
        filedata = pathlib.Path(filepath)
        
        filetype = filedata.suffix
        path = os.path.abspath(filepath)
        
        return Request._post_from_file(
            path,
            filetype,
            data=self.data,
            url_args={"max_bytes": max_bytes, "code_length": code_length, "custom_code": custom_code, "extension": extension},
            max_bytes=max_bytes
        )
    
    def delete_file(self, file_id: str) -> bool:
        """Deletes a file on wants-to.party with the given ID and filetype.
        
        NOTE: you MUST include the filetype! (I.E., ".mp3")"""
        return Request._delete_file(file_id, data=self.data)

    def get_files(self) -> list:
        """Returns a list of file objects relating to your files on wants-to.party
        or `None` if there's no files.
        
        As per a wants-to.party limitation, this only shows the first 1000 files."""
        return Request._get_files(data=self.data)
    
    def get_user(self) -> User:
        """Returns a user object of the user associated with the given API key."""
        return Request._get_user(data=self.data)


def validate_file_extension(extension: str) -> bool:
    """Returns `True` if the file extension provided is allowed 
    on wants-to.party.
    
    Allowed file extensions are:
    png, jpg, gif, mp4, mp3, json, txt, js, html, css, py

    Note that different variations of the same filetype don't work.
    E.G., "jpeg" won't work."""
    if extension.startswith("."):
        extension = extension.replace(".", "")

    allowed = [
        "png",
        "jpg",
        "gif",
        "mp4",
        "mp3",
        "json",
        "txt",
        "js",
        "html",
        "css",
        "py"
    ]
    
    if extension in allowed:
        return True
    
    return False

def _format_arguments(data: dict):
    final = []

    if data.get("max_length"):
        final.append("max_length=" + data["max_length"])
    
    if data.get("custom_code"):
        final.append("custom_code=" + data["custom_code"])
    
    if data.get("code_length"):
        final.append("code_length=" + data["code_length"])
    
    if data.get("extension") != None:
        final.append("extension=" + str(data["extension"]))
    
    arguments = "&".join(final)

    return f"?{arguments}"


import requests
from typing import Union
from aiohttp import ClientSession

from .errors import MaxFileSizeExceeded

class SizeLimit:
    """A class that defines a maximum file size for a file.
    
    This transforms B, KB & MB units into the no. of bytes specified.
    
    To specify the maximum number of bytes, use the kwargs.
    If the maximum file size is exceeded, the MaxFileSizeExceeded exception is raised.
    
    GB+ units not necessary due to wants-to.party's maximum allowed file size of 100MB."""
    def __init__(
        self,
        *,
        mb: float = None,
        kb: float = None,
        b:  float = None
        ):
        self.mb: Union[float, bool] = mb
        self.kb: Union[float, bool] = kb
        self.b:  Union[float, bool] = b

        self.exceeded: bool = False
    
    def _check_bytes(self, all_bytes: int) -> None:
        nbytes = 1024
        b = 0
        kb = 0
        mb = 0

        if self.mb:
            mb = self.mb * (nbytes ** 2)
        
        if self.kb:
            kb = self.kb * (nbytes ** 1)
        
        if self.b:
            b = self.b
        
        total_allowed_bytes = mb + kb + b

        if all_bytes > total_allowed_bytes: # file exceeded max allowed filesize
            self.exceeded = True
            raise MaxFileSizeExceeded("This file exceeded the maximum file size limit.")

class FileSize:
    KBSIZE = 1024
    MBSIZE = 1048576 # in bytes
    def __init__(self, _total: int):
        self._total = _total
    
    def __str__(self) -> str:
        count = self.formatted_bytes()
        fmt = f"{count['mb']}MB, {count['kb']}KB, {count['b']}B"
        return str(fmt)

    @property
    def total_bytes(self) -> int:
        """The number of bytes present in the file."""
        return self._total

    @property
    def total_kb(self) -> int:
        """The number of kilobytes present in the file."""
        return self._total / self.KBSIZE
    
    @property
    def total_mb(self) -> int:
        """The number of megabytes present in the file."""
        return (self._total / self.KBSIZE) / self.KBSIZE
    
    @property
    def formatted_bytes(self) -> dict:
        """The number of MB, KB & B in the file.

        This counts bytes in order of largest unit to smallest. (MB to B)
        If the sum bytes is less than 1 MB, it will calculate number of KB's instead (and so on).

        This method factors in sizes that are on the boundary of the next unit (I.E, 1024 B = 1 KB)

        For example:
        1586230 B -> 1 MB, 525 KB, 54 B"""
        temp = self._total

        mb = 0
        kb = 0
        b = 0

        while temp > 0:
            if temp >= 1048576:
                mb += 1
                temp = temp - 1048576
            elif temp >= 1024:
                kb += 1
                temp = temp - 1024
            else:
                b = temp
                temp = 0
            
        return {"mb": mb, "kb": kb, "b": b}


class User:
    """Represents a wants-to.party user."""
    def __init__(self, data: dict):
        self.data = data

    def __str__(self) -> str:
        return str(self.data.get("user_id"))

    @property
    def user_id(self) -> int:
        """The users ID.
        
        If you're using an oauth session, this is only accessible with the `user.read` scope.
        If you're using your user API key, the `user.read` scope is not necessary."""
        return self.data.get("user_id")
    
    @property
    def userid(self) -> int:
        """(An alias for the user_id property) 
        The users ID.
        
        If you're using an oauth session, this is only accessible with the `user.read` scope.
        If you're using your user API key, the `user.read` scope is not necessary."""
        return self.data.get("user_id")

    @property
    def user_name(self) -> str:
        """The users name.
        
        If you're using an oauth session, this is only accessible with the `user.read` scope.
        If you're using your user API key, the `user.read` scope is not necessary."""
        return self.data.get("user_name")

    @property
    def username(self) -> str:
        """(An alias for the user_name property)
        The users name.
        
        If you're using an oauth session, this is only accessible with the `user.read` scope.
        If you're using your user API key, the `user.read` scope is not necessary."""
        return self.data.get("user_name")

    @property
    def email(self) -> str:
        """The users email.
        
        If you're using an oauth session, this is only accessible with the `user.email` scope.
        If you're using your user API key, the `user.email` scope is not necessary."""
        return self.data.get("email")
    
class File:
    """Represents a wants-to.party file."""
    def __init__(self, data: dict):
        self.data = data

    def __str__(self) -> str:
        return str(self.data.get("url"))

    @property
    def name(self) -> str:
        """The filename and filetype."""
        return self.data.get("name")

    @property
    def extension(self) -> str:
        """The file extension. (does not include the ".")"""
        return self.data.get("extension")

    @property
    def filename(self) -> str:
        """The filename only."""
        return self.data.get("filename")

    @property
    def uploaded_at(self) -> int:
        """The Unix timestamp for when the file was created."""
        return self.data.get("uploaded_at")

    @property
    def url(self) -> str:
        """The full usable URL of the file."""
        return self.data.get("url")

    @classmethod    
    def get_size(self, file_id: str, subdomain: str) -> FileSize:
        """Returns a `FileSize` object  a hosted files' size.
        
        Attributes
        -----------
        file_id - The file ID and filetype of the hosted file.

        subdomain - The the subdomain which the file is hosted with.
        """
        url = f"https://{subdomain}.wants-to.party/{file_id}"

        bcount = len(requests.get(url, stream=True).headers['Content-Length'])

        return FileSize(bcount)
    
    @classmethod
    async def async_get_size(self, file_id: str, subdomain: str) -> FileSize:
        """A non-blocking method that returns a `FileSize` object  a hosted files' size.
        
        Attributes
        -----------
        file_id - The file ID and filetype of the hosted file.

        subdomain - The the subdomain which the file is hosted with.
        """
        url = f"https://{subdomain}.wants-to.party/{file_id}"

        async with ClientSession() as s:
            async with s.get(url) as resp:
                bcount = len(await resp.read())
        
        return FileSize(bcount)


import os
import requests
from aiohttp import ClientSession, FormData

from .errors import BadRequest, _handle_errorcode
from .utils import _format_arguments
from .objects import File, SizeLimit, User

class URL:
    BASE = "https://wants-to.party"
    def __init__(self, path: str, args: dict = None):
        self.path = path
        self.args = args

    def _assemble_url(self):
        args = _format_arguments(self.args) if self.args else ""

        return self.BASE + self.path + args

class Request:
    def _post_from_file(
        file,
        filetype: str,
        *,
        data: dict,
        url_args: dict = None,
        max_bytes: SizeLimit = None
    ):
        key = data.get("key")
        headers = {"Authorization": f"User {key}"}

        if max_bytes:
            max_bytes._check_bytes(os.path.getsize(file))
        
        file_data = open(file, "rb")

        file = {"file": ("file" + filetype, file_data)}

        resp = requests.post(
            URL("/api/user/files", url_args)._assemble_url(),
            headers=headers,
            files=file
        )
        payload = resp.json()

        if resp.status_code != 200:
            return _handle_errorcode(resp.status_code, payload["message"])
        
        return payload["url"]
    
    def _post_from_bytes(
        file,
        filetype: str,
        *,
        data: dict,
        url_args: dict = None,
        max_bytes: SizeLimit = None
    ):
        key = data.get("key")
        headers = {"Authorization": f"User {key}"}

        if max_bytes:
            max_bytes._check_bytes(file.getbuffer().nbytes)
        
        if not filetype.endswith("."):
            filetype = "." + filetype

        filename = f"file{filetype}"

        file = {"file": (filename, file.getvalue())}

        resp = requests.post(
            URL("/api/user/files", url_args)._assemble_url(),
            headers=headers,
            files=file
        )

        payload = resp.json()

        if resp.status_code != 200:
            return _handle_errorcode(resp.status_code, payload["message"])
        
        return payload["url"]

    def _get_user(data: dict):
        key = data.get("key")
        headers = {"Authorization": f"User {key}"}
        resp = requests.get(
            URL("/api/user")._assemble_url(),
            headers=headers
        )
        
        payload = resp.json()

        if resp.status_code != 200:
            return _handle_errorcode(resp.status_code, payload["message"])
        
        return User(payload)
    
    def _get_files(data: dict):
        key = data.get("key")
        headers = {"Authorization": f"User {key}"}
        resp = requests.get(
            URL("/api/user/files")._assemble_url(),
            headers=headers
        )

        payload = resp.json()

        if resp.status_code != 200:
            return _handle_errorcode(resp.status_code, payload["message"])

        files = payload["files"]

        if not files:
            return None
            
        new_files = []
        for file in files:
            new_files.append(File(file))

        return new_files

    def _delete_file(file_id: str, data: dict):
        key = data.get("key")
        headers = {"Authorization": f"User {key}"}
        resp = requests.delete(
            URL(f"/api/user/files/{file_id}")._assemble_url(),
            headers=headers
        )

        payload = resp.json()

        if payload["message"] == "ok":    # lord forgive me
            return True
        
        raise BadRequest("That file does not exist.") # just gonna raise 400 even though its 404.

class AsyncRequest:
    async def _post_from_file(
        file,
        filetype: str,
        *,
        data: dict,
        url_args: dict = None,
        max_bytes: SizeLimit = None
    ):
        key = data.get("key")
        headers = {"Authorization": f"User {key}"}

        if max_bytes:
            max_bytes._check_bytes(os.path.getsize(file))
        
        form = FormData()
        form.add_field("file", open(file, "rb").read(), filename="file" + filetype)
    
        async with ClientSession(headers=headers) as s:
            async with s.post(URL("/api/user/files", url_args)._assemble_url(), data=form) as resp:
                payload = await resp.json()
                
            if resp.status != 200:
                return _handle_errorcode(resp.status, payload["message"])

            return payload["url"]
    
    async def _post_from_bytes(
        file,
        filetype: str,
        *,
        data: dict,
        url_args: dict = None,
        max_bytes: SizeLimit = None
    ):
        key = data.get("key")
        headers = {"Authorization": f"User {key}"}

        if max_bytes:
            max_bytes._check_bytes(file.getbuffer().nbytes)
        
        form = FormData(quote_fields=False)
        filename = f"file{filetype}"
        form.add_field("file", file.getvalue(), filename=filename)
    
        async with ClientSession(headers=headers) as s:
            async with s.post(URL("/api/user/files", url_args)._assemble_url(), data=form) as resp:
                payload = await resp.json()
                
            if resp.status != 200:
                return _handle_errorcode(resp.status, payload["message"])

            return payload["url"]
        
    
    async def _get_user(data: dict):
        key = data.get("key")
        headers = {"Authorization": f"User {key}"}
        async with ClientSession(headers=headers) as s:
            async with s.get(URL("/api/user")._assemble_url()) as resp:
                payload = await resp.json()
                
            if resp.status != 200:
                return _handle_errorcode(resp.status, payload["message"])
            
            return User(payload)
    
    async def _get_files(data: dict):
        key = data.get("key")
        headers = {"Authorization": f"User {key}"}
        async with ClientSession(headers=headers) as s:
            async with s.get(URL("/api/user/files")._assemble_url()) as resp:
                payload = await resp.json()
                
            if resp.status != 200:
                return _handle_errorcode(resp.status, payload["message"])
        
        files = payload["files"]

        if not files:
            return None

        new_files = []
        for file in files:
            new_files.append(File(file))

        return new_files
    
    async def _delete_file(file_id: str, data: dict):
        key = data.get("key")
        headers = {"Authorization": f"User {key}"}
        async with ClientSession(headers=headers) as s:
            async with s.delete(URL(f"/api/user/files/{file_id}")._assemble_url()) as resp:
                payload = await resp.json()
                
            if resp.status != 200:
                return _handle_errorcode(resp.status, payload["message"])
        
            return True



from setuptools import find_packages, setup

VERSION = "2.1.1"
DESCRIPTION = "An API wrapper for the wants-to.party API."

long_desc = ""
with open(".github/README.md", "r") as f:
    long_desc = f.read()

setup(
    name="wantstoparty",
    version=VERSION,
    author="acatiadroid",
    license="The MIT License (MIT)",
    author_email="<acatia@mail.com>",
    url="https://github.com/acatiadroid/wantstoparty",
    description=DESCRIPTION,
    long_description=long_desc,
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=["docs", "examples", "build", "dist", "*.egg-info"]),
    install_requires=["aiohttp", "requests"],
    keywords=["python", "api wrapper"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)